(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREER SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Wf.
Require Import Wellfounded.
Require Import Relations.

Require Import rel_utils.
Require Import list_utils.
Require Import good_base.
Require Import wft.
Require Import af_t.
Require Import bar_t.

Set Implicit Arguments.

Section af_bar_t.

  Variable (X : Type).

  Implicit Type (R : X -> X -> Prop).
 
  Section bar_t_good_af.

    Lemma bar_good_af_t R ll : bar_t (good R) ll -> af_t (R lrlift ll).
    Proof.
      induction 1 as [ ll Hll | ll Hll IH ] using bar_t_rect.
      
      apply in_af_t0.
      induction Hll as [ ll a b Hll Hab | ll a Hll IH ] .
      intros; simpl; right; apply lift_rel_list_In with (1 := Hab); auto.
      intros; simpl; left; auto.
      
      apply in_af_t1, IH.
    Qed.

    Let af_bar_good_rec R : af_t R -> forall S ll, R inc2 lift_rel_list S ll -> bar_t (good S) ll.
    Proof.
      induction 1 as [ R HR | R HR IHR ] using af_t_rect; intros S ll H.
      apply in_bar_t1; intros u.
      apply in_bar_t1; intros v.
      apply in_bar_t0.
      cutrewrite (v::u::ll = (v::u::nil)++ll); auto.
      apply good_lift_rel_list.
      apply good_inc with (1 := H).
      constructor 1 with u; auto; left; auto.
      apply in_bar_t1; intros a.
      apply (IHR a _ (a::ll)).
      simpl; unfold lift_rel.
      intros ? ? [ | ]; [ left | right ]; auto.
    Qed.

    Lemma af_bar_good_t R ll : af_t (R lrlift ll) -> bar_t (good R) ll.
    Proof.
      intros H; apply af_bar_good_rec with (1 := H); auto.
    Qed.

    Theorem af_t_bar_t R : af_t R -> bar_t (good R) nil.
    Proof.
      apply af_bar_good_t with (ll := nil).
    Qed.

    Theorem bar_t_af_t R : bar_t (good R) nil -> af_t R.
    Proof.
      apply bar_good_af_t.
    Qed.

  End bar_t_good_af.

End af_bar_t.

Fact bar_t_good_le : bar_t (good le) nil.
Proof.
  apply af_t_bar_t, le_af_t.
Qed.

Fact bar_t_ramsey X Y R S : bar_t (@good X R) nil 
                         -> bar_t (@good Y S) nil 
                         -> bar_t (good (rel_prod R S)) nil.
Proof.
  intros H1 H2.
  apply af_t_bar_t.
  apply af_t_prod; apply bar_t_af_t; auto.
Qed.
