(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Classical.

Require Import list_utils. (* for pfx_rev *)
Require Import list_dep_choice.

Section bar_induction.

  Variables (X : Type) (P : list X -> Prop).

  Hypothesis Pdec : forall ll, P ll \/ ~ P ll.
  Hypothesis Pfin : forall f, exists n, P (pfx_rev f n).

  Variable Q : list X -> Prop.
    
  Hypothesis Q_0 : forall ll, P ll -> Q ll.
  Hypothesis Q_1 : forall ll, (forall x, Q (x::ll)) -> Q ll.
  
  (* We need Classical Reasonning + Dependent Choice to prove Bar induction *)    
    
  Let bar_next ll : ~ Q ll -> exists x, ~ Q (x::ll).
  Proof.
    intros H.
    apply not_all_ex_not.
    contradict H; apply Q_1; auto.
  Qed.

  Theorem bar_induction : Q nil.
  Proof.
    apply NNPP.
    intros H.
    destruct (functional_list_dep_choice _ (fun ll => ~ Q ll)) as (f & Hf); auto.
    intros; apply classic.
    destruct (Pfin f) as (n & H1).
    apply (Hf n).
    apply Q_0; auto.
  Qed.

End bar_induction.

  