(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import rel_utils.
Require Import list_utils.

Require Import bar.
Require Import bars.

Set Implicit Arguments.

Section fan.

  (* That proof of the FAN theorem is derived from Daniel Fridlender's paper *)

  Notation mono := ((fun X (P : list X -> Prop) => forall x l, P l -> P (x::l)) _).

  Variables (X : Type) (P : list X -> Prop) (HP : mono P).
  
  Let P_mono l m : P m -> P (l++m).
  Proof.
    induction l; simpl; auto.
  Qed.
  
  Let V u := fun lw => Forall (fun v => P (v++u)) (list_fan lw).

  Let V_mono u : mono (V u).
  Proof.
    unfold V; intros w lw H.
    apply list_fan_Forall.
    rewrite Forall_forall; intros x _.
    revert H; apply Forall_impl.
    intro; apply HP.
  Qed.

  Let bar_V u : bar P u -> bar (V u) nil.
  Proof.    
    induction 1 as [ u Hu | u Hu IHu ].
    
    constructor 1; unfold V.
    simpl. 
    constructor; simpl; auto.
    
    constructor 2; intros w.
    induction w as [ | a w IHw ].
    
    constructor 1.
    unfold V; simpl.
    constructor.

    specialize (IHu a).
    rewrite bar_lift in IHw.
    unfold V in IHw.
    generalize (@V_mono (a::u)); intros HV1.
    assert (mono (fun lw => V u (lw++w::nil))) as HV2.
      intros x ll; simpl; apply V_mono.
    generalize (bar_intersection HV1 HV2 IHu IHw); intros HV3; clear HV1 HV2.
    rewrite bar_lift.
    revert HV3; apply bar_inc.

    intros x; unfold V.
    intros [ H1 H2 ].
    rewrite Forall_perm with (1 := list_fan_middle_cons x a w nil).
    apply Forall_app; split; auto.
    rewrite list_fan_sg_right.
    rewrite Forall_map.
    revert H1; apply Forall_impl.
    intro; rewrite app_ass; auto.
  Qed.

  Theorem fan_on_list : bar P nil -> bar (fun ll => Forall P (list_fan ll)) nil.
  Proof.
    intros H.
    generalize (bar_V H).
    apply bar_inc.
    intros lw; unfold V.
    apply Forall_impl.
    intro; rewrite <- app_nil_end; auto.
  Qed.

End fan.

Section fans.

  Variables (X : Type) (P : X -> Prop) (Q : list X -> Prop) (HP : forall x l, P x -> Q l -> Q (x::l)).

  Let f (ll : list (sig P)) : sig (Forall P).
  Proof.
    revert ll.
    refine (fix loop ll := match ll with
       | nil              => exist _ nil _
       | (exist x Hx)::mm => exist _ (x::proj1_sig (loop mm)) _
    end).
    constructor.
    constructor.
    apply Hx.
    apply (proj2_sig (loop mm)).
  Defined.

  Let g : sig (Forall P) -> list (sig P).
  Proof.
    intros (ll & Hll); revert ll Hll.
    refine (fix loop ll := match ll return Forall P ll -> _ with
      | nil   => fun _   => nil
      | x::mm => fun Hll => _
    end).
    apply Forall_cons_inv in Hll.
    destruct Hll as [ H1 H2 ].
    apply (exist _ x H1::loop mm H2).
  Defined.
  
  Let f_o_g x : proj1_sig (f (g x)) = proj1_sig x.
  Proof.
    destruct x as (ll & Hll).
    simpl.
    induction ll as [ | x ll IH ]; simpl; auto.
    destruct (Forall_cons_inv P x ll) as (H1 & H2); simpl.
    destruct (H1 Hll); simpl.
    f_equal; auto.
  Qed.

  Let In_f u ll : In u (proj1_sig (f ll)) -> exists H, In (exist _ u H) ll.
  Proof.
    revert u.
    induction ll as [ | (x,Hx) ll Hll ]; simpl; intros u.
    intros [].
    intros [ H | H ]; subst.
    exists Hx; auto.
    specialize (Hll _ H).
    destruct Hll as (Hu & Hll).
    exists Hu; auto.
  Qed.

  Fact fans_on_list : bars P Q nil -> bars (Forall P) (fun ll : list (list X) => Forall Q (list_fan ll)) nil.
  Proof.
    do 2 rewrite <- bar_eq_bars.
    intros H.
    apply fan_on_list in H.
    fold (map f nil).
    revert H.
    generalize (@nil (list (sig P))).
    induction 1 as [ ll Hll | ll Hll IH ].

    constructor 1.    
    rewrite Forall_forall.
    intros x Hx.
    apply list_fan_spec in Hx.
    rewrite map_map in Hx.
    rewrite Forall2_map_right in Hx.
    rewrite Forall_forall in Hll.
    assert (exists xl : list (sig P), map (@proj1_sig _ _) xl = x /\ In xl (list_fan ll)).
      clear Hll.
      induction Hx as [ | u v ll mm Hull H0 (xl & H3 & H4) ].
      exists nil; split; simpl; auto.
      destruct (In_f _ _ Hull) as (Hu & H5).
      exists (exist _ u Hu::xl).
      simpl; split. 
      f_equal; tauto.
      apply list_prod_spec.
      exists (exist _ u Hu), xl; auto.
    destruct H as (xl & H1 & H2).
    rewrite <- H1.    
    apply Hll; auto.
    
    constructor 2; intros (a & Ha).
    generalize (f_o_g (exist _ a Ha)); intros H0.
    generalize (IH (g (exist _ a Ha))).
    apply bar_subst_eq with (mm := nil).
    intros mm kk.
    do 2 rewrite map_app.
    unfold map at 4; fold map.
    rewrite <- f_o_g.
    auto.
    
    intros (x & Hx) ll; simpl; auto.
  Qed.

End fans.


