(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.
Require Import Wellfounded.

Require Omega.

Require Import rel_utils.
Require Import list_utils.

Require Import af.
Require Import af_t.

Require Import abtree.

Set Implicit Arguments.

Section higman_abt_aux0.

  Variables (X Y Z : Type) 
            (RX : X -> X -> Prop)
            (RY : Y -> Y -> Prop)
            (RZ : Z -> Z -> Prop).
           
  Fact higman_abt_aux00 x : af (embed_abtree (RX rlift x) RY RZ) 
                         -> af (embed_abtree RX RY RZ rlift abt_leaf x).
  Proof.
    apply af_inc.
    induction 1 as [ ? ? [|] 
                   | ? ? ? ? ? ? [|]
                   | ? ? ? ? ? ? ? ? [|] ? [|]
                   | ? ? ? ? [|]
                   | ? ? ? ? ? [|]
                   | ? ? ? ? ? [|]
                   ].
    left; constructor 1; auto.
    right; constructor 1; auto.

    left; constructor 2; auto.
    right; constructor 4; auto.
    
    left; constructor 3; auto.
    right; constructor 6; auto.
    right; constructor 5; auto.
    right; constructor 6; auto.
    
    left; constructor 4; auto.
    right; auto.
    
    left; constructor 5; auto.
    right; auto.
    
    left; constructor 6; auto.
    right; auto.
  Qed.

  Fact higman_abt_aux00_t x : af_t (embed_abtree (RX rlift x) RY RZ) 
                           -> af_t (embed_abtree RX RY RZ rlift abt_leaf x).
  Proof.
    apply af_t_inc.
    induction 1 as [ ? ? [|] 
                   | ? ? ? ? ? ? [|]
                   | ? ? ? ? ? ? ? ? [|] ? [|]
                   | ? ? ? ? [|]
                   | ? ? ? ? ? [|]
                   | ? ? ? ? ? [|]
                   ].
    left; constructor 1; auto.
    right; constructor 1; auto.

    left; constructor 2; auto.
    right; constructor 4; auto.
    
    left; constructor 3; auto.
    right; constructor 6; auto.
    right; constructor 5; auto.
    right; constructor 6; auto.
    
    left; constructor 4; auto.
    right; auto.
    
    left; constructor 5; auto.
    right; auto.
    
    left; constructor 6; auto.
    right; auto.
  Qed.

  Fact higman_abt_aux01 x : full RX 
                         -> af (embed_abtree RX RY RZ rlift abt_leaf x).
  Proof.
    intros HX.
    apply in_af_0.
    intros t1 t2; right.
    destruct (abt_leaf_exists t1) as (y & Hy).
    apply sabt_leaf_inv.
    exists y; auto.
  Qed.

  Fact higman_abt_aux01_t x : full RX 
                           -> af_t (embed_abtree RX RY RZ rlift abt_leaf x).
  Proof.
    intros HX.
    apply in_af_t0.
    intros t1 t2; right.
    destruct (abt_leaf_exists t1) as (y & Hy).
    apply sabt_leaf_inv.
    exists y; auto.
  Qed.

End higman_abt_aux0.

Section higman_abt_aux10.

  Variables (X Y Z : Type) 
            (RX : X -> X -> Prop)
            (RY : Y -> Y -> Prop)
            (RZ : Z -> Z -> Prop)
            (y : Y) (t : abtree X Y Z).

  Hypothesis HRX : af RX.
  Hypothesis HRY : full RY.
  Hypothesis IHY : forall X' RX', @af X' RX' -> af (embed_abtree RX' ER RZ).
  Hypothesis Ht  : af (embed_abtree RX RY RZ rlift t).
  
  Fact higman_abt_aux10 : af (embed_abtree RX RY RZ rlift abt_cons y t).
  Proof.
    apply abt_ev2_af, IHY; auto.
    apply abt_rel2_af; auto.
  Qed.

  Hypothesis HRX_t : af_t RX.
  Hypothesis IHY_t : forall X' RX', @af_t X' RX' -> af_t (embed_abtree RX' ER RZ).
  Hypothesis Ht_t  : af_t (embed_abtree RX RY RZ rlift t).

  Fact higman_abt_aux10_t : af_t (embed_abtree RX RY RZ rlift abt_cons y t).
  Proof.
    apply abt_ev2_af_t, IHY_t; auto.
    apply abt_rel2_af_t; auto.
  Qed.
 
End higman_abt_aux10.

Section higman_abt_aux11.

  Variables (X Y Z : Type) 
            (RX : X -> X -> Prop)
            (RY : Y -> Y -> Prop)
            (RZ : Z -> Z -> Prop)
            (y : Y) (t : abtree X Y Z).

  Hypothesis HRX : af RX.
  Hypothesis HRY : af RY.
  Hypothesis IHY : forall X' RX', @af X' RX' -> af (embed_abtree RX' (RY rlift y) RZ).
  Hypothesis Ht : af (embed_abtree RX RY RZ rlift t).
  
  Fact higman_abt_aux11 : af (embed_abtree RX RY RZ rlift abt_cons y t).
  Proof.
    apply abt_ev1_af, IHY; auto.
    apply abt_rel1_af; auto.
  Qed.

  Hypothesis HRX_t : af_t RX.
  Hypothesis HRY_t : af_t RY.
  Hypothesis IHY_t : forall X' RX', @af_t X' RX' -> af_t (embed_abtree RX' (RY rlift y) RZ).
  Hypothesis Ht_t  : af_t (embed_abtree RX RY RZ rlift t).
  
  Fact higman_abt_aux11_t : af_t (embed_abtree RX RY RZ rlift abt_cons y t).
  Proof.
    apply abt_ev1_af_t, IHY_t; auto.
    apply abt_rel1_af_t; auto.
  Qed.
 
End higman_abt_aux11.

Section higman_abt_aux20.

  Variables (X Y Z : Type) 
            (RX : X -> X -> Prop)
            (RY : Y -> Y -> Prop)
            (RZ : Z -> Z -> Prop)
            (z : Z) (t1 t2 : abtree X Y Z).

  Hypothesis HRX : af RX.
  Hypothesis HRY : af RY.
  Hypothesis HRZ : full RZ.
  Hypothesis IHZ : forall X' RX' Y' RY', @af X' RX' -> @af Y' RY' -> af (embed_abtree RX' RY' ER).
  Hypothesis Ht1 : af (embed_abtree RX RY RZ rlift t1).
  Hypothesis Ht2 : af (embed_abtree RX RY RZ rlift t2).
  
  Lemma higman_abt_aux20 : af (embed_abtree RX RY RZ rlift abt_pair z t1 t2).
  Proof.
    apply abt_ev4_af, IHZ; auto.
    apply abt_rel4_af; auto.
  Qed.

  Hypothesis HRX_t : af_t RX.
  Hypothesis HRY_t : af_t RY.
  Hypothesis HRZ_t : full RZ.
  Hypothesis IHZ_t : forall X' RX' Y' RY', @af_t X' RX' -> @af_t Y' RY' -> af_t (embed_abtree RX' RY' ER).
  Hypothesis Ht1_t : af_t (embed_abtree RX RY RZ rlift t1).
  Hypothesis Ht2_t : af_t (embed_abtree RX RY RZ rlift t2).
  
  Lemma higman_abt_aux20_t : af_t (embed_abtree RX RY RZ rlift abt_pair z t1 t2).
  Proof.
    apply abt_ev4_af_t, IHZ_t; auto.
    apply abt_rel4_af_t; auto.
  Qed.

End higman_abt_aux20.

Section higman_abt_aux21.

  Variables (X Y Z : Type) 
            (RX : X -> X -> Prop)
            (RY : Y -> Y -> Prop)
            (RZ : Z -> Z -> Prop)
            (z : Z) (t1 t2 : abtree X Y Z).

  Hypothesis HRX : af RX.
  Hypothesis HRY : af RY.
  Hypothesis HRZ : af RZ.
  Hypothesis IHZ : forall X' RX' Y' RY', @af X' RX' -> @af Y' RY' -> af (embed_abtree RX' RY' (RZ rlift z)).
  Hypothesis Ht1 : af (embed_abtree RX RY RZ rlift t1).
  Hypothesis Ht2 : af (embed_abtree RX RY RZ rlift t2).
  
  Lemma higman_abt_aux21 : af (embed_abtree RX RY RZ rlift abt_pair z t1 t2).
  Proof.
    apply abt_ev3_af, IHZ; auto.
    apply abt_rel3_af; auto.
  Qed.
  
  Hypothesis HRX_t : af_t RX.
  Hypothesis HRY_t : af_t RY.
  Hypothesis HRZ_t : af_t RZ.
  Hypothesis IHZ_t : forall X' RX' Y' RY', @af_t X' RX' -> @af_t Y' RY' -> af_t (embed_abtree RX' RY' (RZ rlift z)).
  Hypothesis Ht1_t : af_t (embed_abtree RX RY RZ rlift t1).
  Hypothesis Ht2_t : af_t (embed_abtree RX RY RZ rlift t2).
  
  Lemma higman_abt_aux21_t : af_t (embed_abtree RX RY RZ rlift abt_pair z t1 t2).
  Proof.
    apply abt_ev3_af_t, IHZ_t; auto.
    apply abt_rel3_af_t; auto.
  Qed.
 
End higman_abt_aux21.

Fact abtree_no_leaf Y Z (t : abtree Empty_set Y Z) : False.
Proof.
  destruct (abt_leaf_exists t) as ( [] & _) .
Qed.

Fact af_ER : af ER.
Proof.
  constructor 1; intros [].
Qed.

Fact af_t_ER : af_t ER.
Proof.
  apply in_af_t0; intros [].
Qed.

Section higman_abt.

  Hint Resolve higman_abt_aux00 higman_abt_aux01 
               higman_abt_aux10 higman_abt_aux11 
               higman_abt_aux20 higman_abt_aux21
               in_af_1 in_af_0 af_ER.

  Ltac rem_empty := 
    try match goal with 
      | [ x: Empty_set |- _ ]            => destruct x
      | [ t: abtree Empty_set _ _ |- _ ] => destruct (abtree_no_leaf t) 
    end.

  Ltac my_crush := let x := fresh in constructor 2; intros x; rem_empty; induction x; rem_empty; auto.
    
  Theorem higman_abt X Y Z R S T : @af X R -> @af Y S -> @af Z T -> af (embed_abtree R S T).
  Proof.
    intros HR HS HT; revert Z T HT Y S HS X R HR.
    do 3 (induction 1 using af_dep_ind); my_crush.
  Qed.
       
End higman_abt.

Check higman_abt.
Print Assumptions higman_abt.

Section higman_abt_t.

  Hint Resolve higman_abt_aux00_t higman_abt_aux01_t 
               higman_abt_aux10_t higman_abt_aux11_t 
               higman_abt_aux20_t higman_abt_aux21_t
               in_af_t1 in_af_t0 af_t_ER.

  Ltac rem_empty := 
    try match goal with 
      | [ x: Empty_set |- _ ]            => destruct x
      | [ t: abtree Empty_set _ _ |- _ ] => destruct (abtree_no_leaf t) 
    end.

  Ltac my_crush x := apply in_af_t1; intros x; rem_empty; induction x; rem_empty; auto.
    
  Theorem higman_abt_t X Y Z R S T : @af_t X R -> @af_t Y S -> @af_t Z T -> af_t (embed_abtree R S T).
  Proof.
    intros HR HS HT; revert Z T HT Y S HS X R HR.
    do 3 (induction 1 using af_t_dep_rect); my_crush t.
  Qed.
 
End higman_abt_t.

Check higman_abt_t.
Print Assumptions higman_abt_t.

