(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Omega.

Set Implicit Arguments.

Notation lsum := (fold_right plus 0).
 
Fact lsum_prop x ll : In x ll -> x <= lsum ll.
Proof.
  induction ll as [ | y ll IH ]; simpl.
  intros [].
  intros [ | H ]; subst.
  omega.
  specialize (IH H); omega.
Qed.

Fixpoint list_n n := match n with 0 => nil | S n => n::list_n n end.
  
Fact list_n_prop n x : x < n <-> In x (list_n n).
Proof.
  revert x; induction n as [ | ? IH ]; intros ?.
  split. 
  omega.
  intros [].
  simpl; rewrite <- IH.
  split; omega.
Qed.

Fact list_n_length x : length (list_n x) = x.
Proof. induction x; simpl; f_equal; auto. Qed.

Fact largest_nat_prefix n (P : nat -> Prop) : (forall i, i <= n -> { P i } + { ~ P i })
                                           -> P 0
                                           -> { i | i <= n 
                                                 /\ P i 
                                                 /\ (P (S i) -> n <= i) 
                                                 /\ forall j, j <= i -> P j }.
Proof.
  revert P.
  induction n as [ | n IHn ]; intros P HP H.

  exists 0.
  repeat split; auto.
  intros j Hj.
  cutrewrite (j = 0); auto; omega.
  
  destruct (HP 1) as [ H1 | H1 ]; try omega.

  destruct (IHn (fun x => P (S x))) as (i & H3 & H4 & H5 & H6); auto.
  intros; apply  HP; omega.
  exists (S i); repeat split; auto. 
  omega.
  intros H'; apply H5 in H'; omega.
  intros [ | j ]; auto.
  intros H7; apply H6; omega.
  
  exists 0; repeat split; auto.
  omega.
  tauto.
  intros j Hj; cutrewrite (j=0); auto; omega.
Qed.
 