(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import sublist.
Require Import subword.

Require Import tree.

Require Export tree_emb_product.
Require Export tree_emb_upto.
Require Export tree_emb_homeo.

Set Implicit Arguments.

Section emb_tree_higman_emb_tree_upto.

  (* on a set P of trees where nodes of arities over k are not autorized, 
     the relations emb_ntree R and emb_tree_upto k R are equivalent *)

  Variable (X : Type) (P : nat -> X -> Prop) (R : nat -> X -> X -> Prop) (k : nat). 
  
  Hypothesis HP : forall n x, k <= n -> ~ P n x.

  Fact emb_tree_higman_emb_tree_upto s t : wfptree P s -> wfptree P t -> (emb_tree_higman R s t <-> emb_tree_upto k R s t).
  Proof.
    intros H1 H2; split; revert s t H1 H2.

    apply emb_wfptree_higman_ind.

    intros s t x ll Hs Hll H1 H2 H3.
    apply in_emb_tree_upto_0 with t; auto.

    intros x y ll mm Hll Hmm H1 H2 H3.
    destruct (le_lt_dec k (length ll)).
    rewrite wfptree_fix in Hll.
    apply proj1, HP in Hll; tauto.
    apply in_emb_tree_upto_1; auto.
    
    apply emb_wfptree_upto_ind.
    
    intros s t x ll Hs Hll H1 H2 H3.
    constructor 1 with t; auto.
    
    intros x y ll mm Hll Hmm H1 H2 _ H3.
    constructor 2; auto.
    
    intros x y ll mm Hll _ H1 _ _.
    rewrite wfptree_fix in Hll.
    apply proj1, HP in Hll; tauto.
  Qed.

  Fact emb_tree_higman_emb_tree_upto_eq : emb_tree_higman R <# wfptree P #> ~eq2 emb_tree_upto k R <# wfptree P #>.
  Proof.
    split; intros (s & Hs) (t & Ht);
      unfold rel_restr; simpl;
      apply emb_tree_higman_emb_tree_upto; auto.
  Qed.

End emb_tree_higman_emb_tree_upto.

Section emb_tree_product_upto_homeo.

  Variable (X : Type) (R : X -> X -> Prop).

  Fact emb_tree_upto_homeo k t1 t2 :  emb_tree_upto k (fun _ => R) t1 t2
                                   -> emb_tree_homeo R t1 t2.
  Proof.
    induction 1 as [ t1 t2 a ll H1 H2 H3
                   | a b ll mm H1 H2 H3
                   | a b ll mm H1 H2 H3 ] using emb_tree_upto_ind.
    apply in_emb_tree_homeo_0 with t2; auto.
    apply in_emb_tree_homeo_1; auto.
    apply subword_Forall2; auto.
    apply in_emb_tree_homeo_1; auto.
  Qed.

  Fact emb_tree_homeo_upto_eq t1 t2 : emb_tree_homeo R t1 t2 <-> emb_tree_upto 0 (fun _ => R) t1 t2.
  Proof.
    split.
    
    induction 1 as [ t1 t2 a ll H1 H2 H3 
                   | a b ll mm H1 H2 H3 ] using emb_tree_homeo_ind.
    apply in_emb_tree_upto_0 with t2; auto.
    apply in_emb_tree_upto_2; auto; omega.
    
    apply emb_tree_upto_homeo.
  Qed.

  Fact emb_tree_product_emb_tree_homeo t1 t2 : emb_tree_product R t1 t2 -> emb_tree_homeo R t1 t2.
  Proof.
    induction 1 as [ ? t | ].
    apply in_emb_tree_homeo_0 with t; auto.
    apply in_emb_tree_homeo_1; auto.
    apply subword_Forall2; auto.
  Qed.

End emb_tree_product_upto_homeo.
